from __future__ import annotations
import re
import numpy as np
import PIL
import os

from config import *


def assert_path(path:str, variable_name:str) -> None:
    assert_types([path, variable_name], [str, str], ["variable_name", "path"])
    if not os.path.exists(path):
        raise ValueError(f"Received bad path: `{variable_name}`=`{path}`")

def assert_image_type(image_source, variable_name:str) -> None:
    assert_type(variable_name, str, variable_name)
    if not TORCH_FOUND:
        assert_in(type(image_source), [np.ndarray, PIL.Image.Image, str], variable_name)
    else:
        assert_in(type(image_source), [np.ndarray, torch.Tensor, PIL.Image.Image, str], variable_name)

def assert_valid_image(image: ndarray) -> None:
    """ Try and ensure that `image` is a legitimate formatted numpy image """
    assert_type(image, ndarray, 'image')

    # Check shape
    assert len(image.shape) in [2, 3], f"Expected the shape of the image to be in [2,3], but received `{len(image.shape)=}`"
    is_greyscale = (len(image.shape) == 2) or (len(image.shape) == 3 and (image.shape[2] == 1))
    if len(image.shape) != 2:
        assert image.shape[-1] in [1, 3, 4], f"Expected to find a greyscale image, RGB or RGBA, but found: `{image.shape=}`"
    if (len(image.shape) != 3) and (not is_greyscale):
        raise ValueError(f"Expected 3 dimensional shape for non-greyscale images (dim1 x dim2 x channels). "
                         f"Received shape `{image.shape}`")

    # Check data type
    if image.dtype != 'uint8':
        extra = ""
        if 'f' in str(image.dtype):
            extra = " If `image` is in 0.0-1.0 float format, try casting it with: `image = (image * 255).astype(np.uint8)`"
        raise TypeError(f"Expected `image` to be of dtype `uint8`, but received `{image.dtype}`.{extra}")

    # Check dimensions
    if image.shape[0] < 3:
        raise ValueError(f"Expected an image with a height of at least 3 pixels, but received `{image.shape[0]=}`")
    if image.shape[1] < 3:
        raise ValueError(f"Expected an image with a width of at least 3 pixels, but received `{image.shape[1]=}`")

    # Check pixel range
    if np.min(image) < 0:
        raise ValueError("Detected pixel value below 0. Consider using `numpy.clip` to get rid of them")
    if np.min(image) > 255:
        raise ValueError("Detected pixel value above 255. Consider using `numpy.clip` to get rid of them")

def is_valid_url(url:str) -> bool:
    """
    Attempt to check if `url` is valid.
    This function is taken more or less directly from the wonderful validators library - https://github.com/kvesteri/validators.
    All credit goes to them.
    """
    assert_type(url, str, "url")
    ip_middle_octet = r"(?:\.(?:1?\d{1,2}|2[0-4]\d|25[0-5]))"
    ip_last_octet = r"(?:\.(?:0|[1-9]\d?|1\d\d|2[0-4]\d|25[0-5]))"

    regex = re.compile(
        r"^"
        r"(?:(?:https?|ftp)://)"
        r"(?:[-a-z\u00a1-\uffff0-9._~%!$&'()*+,;=:]+"
        r"(?::[-a-z0-9._~%!$&'()*+,;=:]*)?@)?"
        r"(?:"
        r"(?P<private_ip>"
        r"(?:(?:10|127)" + ip_middle_octet + r"{2}" + ip_last_octet + r")|"
        r"(?:(?:169\.254|192\.168)" + ip_middle_octet + ip_last_octet + r")|"
        r"(?:172\.(?:1[6-9]|2\d|3[0-1])" + ip_middle_octet + ip_last_octet + r"))"
        r"|"
        r"(?P<private_host>"
        r"(?:localhost))"
        r"|"
        r"(?P<public_ip>"
        r"(?:[1-9]\d?|1\d\d|2[01]\d|22[0-3])"
        r"" + ip_middle_octet + r"{2}"
        r"" + ip_last_octet + r")"
        r"|"
        r"\[("
        r"([0-9a-fA-F]{1,4}:){7,7}[0-9a-fA-F]{1,4}|"
        r"([0-9a-fA-F]{1,4}:){1,7}:|"
        r"([0-9a-fA-F]{1,4}:){1,6}:[0-9a-fA-F]{1,4}|"
        r"([0-9a-fA-F]{1,4}:){1,5}(:[0-9a-fA-F]{1,4}){1,2}|"
        r"([0-9a-fA-F]{1,4}:){1,4}(:[0-9a-fA-F]{1,4}){1,3}|"
        r"([0-9a-fA-F]{1,4}:){1,3}(:[0-9a-fA-F]{1,4}){1,4}|"
        r"([0-9a-fA-F]{1,4}:){1,2}(:[0-9a-fA-F]{1,4}){1,5}|"
        r"[0-9a-fA-F]{1,4}:((:[0-9a-fA-F]{1,4}){1,6})|"
        r":((:[0-9a-fA-F]{1,4}){1,7}|:)|"
        r"fe80:(:[0-9a-fA-F]{0,4}){0,4}%[0-9a-zA-Z]{1,}|"
        r"::(ffff(:0{1,4}){0,1}:){0,1}"
        r"((25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])\.){3,3}"
        r"(25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])|"
        r"([0-9a-fA-F]{1,4}:){1,4}:"
        r"((25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])\.){3,3}"
        r"(25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])"
        r")\]|"
        r"(?:(?:(?:xn--)|[a-z\u00a1-\uffff\U00010000-\U0010ffff0-9]-?)*"
        r"[a-z\u00a1-\uffff\U00010000-\U0010ffff0-9]+)"
        r"(?:\.(?:(?:xn--)|[a-z\u00a1-\uffff\U00010000-\U0010ffff0-9]-?)*"
        r"[a-z\u00a1-\uffff\U00010000-\U0010ffff0-9]+)*"
        r"(?:\.(?:(?:xn--[a-z\u00a1-\uffff\U00010000-\U0010ffff0-9]{2,})|"
        r"[a-z\u00a1-\uffff\U00010000-\U0010ffff]{2,}))"
        r")"
        r"(?::\d{2,5})?"
        r"(?:/[-a-z\u00a1-\uffff\U00010000-\U0010ffff0-9._~%!$&'()*+,;=:@/]*)?"
        r"(?:\?\S*)?"
        r"(?:#\S*)?"
        r"$",
        re.UNICODE | re.IGNORECASE
    )

    pattern = re.compile(regex)
    result = pattern.match(url)
    return result is not None


def assert_type(to_check, expected_type, variable_name:str, allow_none:bool=False) -> None:
    """
    Check object against expected type

    :param to_check: Object for type check
    :param expected_type: Expected type of `to_check`
    :param variable_name: The name of `to_check` that will be displayed in error messages
    :param allow_none: Weather or not None is acceptable
    """

    if not isinstance(allow_none, bool):
        error_message = f"Expected `allow_None` to by of type bool, but received `{type(allow_none)=}`"
        raise TypeError(error_message)
    if not isinstance(variable_name, str):
        error_message = f"Expected `variable_name` to by of type str, but received `{type(variable_name)=}`"
        raise TypeError(error_message)
    if (to_check is None) and (expected_type is None): # TODO: Test this
        error_message = f"`None` is not a valid type. If you're trying to check if `type(to_check) == None` try setting `expected_type=type(None)` instead."
        raise TypeError(error_message)


    try:
        is_ok = isinstance(to_check, expected_type)
    except:
        error_message = f"Failed to execute `isinstance({variable_name}, {expected_type})`, " \
                        f"most likely because `{expected_type}` is not valid for type checking"
        raise TypeError(error_message)
    if allow_none:
        is_ok = (to_check is None) or is_ok

    if not is_ok:
        error_message = f"Expected type `type({variable_name})={type(to_check)}` to be of type `{str(expected_type)}`"
        raise TypeError(error_message)


def assert_types(to_check:list, expected_types:list, variable_names:List[str], allow_nones:list[int|bool]=None) -> None:
    """
    Check list of values against expected types

    :param to_check: List of values for type check
    :param expected_types: Expected types of `to_check`
    :param variable_names: The name of `to_check` that will be displayed in error messages
    :param allow_nones: list of booleans or 0/1
    """

    # Checks
    assert_type(to_check, list, "to_check")
    assert_type(expected_types, list, "expected_types")
    assert_type(variable_names, list, "variable_names")
    assert_type(allow_nones, list, "allow_nones", allow_none=True)
    if len(to_check) != len(expected_types):
        error_message = "length mismatch between `{to_check_values}` and `{expected_types}`"
        raise ValueError(error_message)

    # If `allow_nones` is None all values are set to False.
    if allow_nones is None:
        allow_nones = [False]*len(to_check)
    else:
        if not (len(variable_names) == len(allow_nones) == len(to_check)):
            raise ValueError(f"Expected equal lengths, but found: `{len(to_check)=}`, `{len(allow_nones)=}` and `{len(variable_names)=}`")
        for i, element in enumerate(allow_nones):
            if element in [0, 1]:
                allow_nones[i] = bool(element == 1) # the `== 1` is just to allow for zeros as False and ones as True
            else:
                raise ValueError(f"`{allow_nones=}` may only contain [False, True, 0, 1]")

    # check if all elements are of the correct type
    for v, t, n, a in zip(to_check, expected_types, variable_names, allow_nones):
        assert_type(to_check=v, expected_type=t, variable_name=n, allow_none=a)


def assert_list_types(to_check:Union[list, tuple], expected_type, variable_name:str, allow_none:bool=False) -> None:
    assert_type(to_check, (list, tuple), variable_name)
    for i, item in enumerate(to_check):
        assert_type(item, expected_type, f"{variable_name}[{i}]", allow_none)

def assert_dict_types(dict_name:str, dict_to_check:dict, names:List[str], types:list, must_be_present:list=None, allow_nones:list=None, eval_expr:List[str]=None, assert_dict_lte:bool=True) -> None:
    # Input checks
    assert_types([dict_name, dict_to_check, names, types, must_be_present, allow_nones, eval_expr, assert_dict_lte],
                 [str, dict, list, list, list, list, list, bool],
                 ["dict_name", "dict_to_check", "names", "types", "must_be_present", "allow_nones", "eval_expr", "assert_dict_lte"],
                 [0,0,0,0,1,1,1,0])
    assert_list_types(names, str, "names")

    # Check optionals and create dummies if None (avoid issues with different scenarios below that way)
    if must_be_present is not None: assert_list_types(must_be_present, (bool, int), "must_be_present")
    else: must_be_present = [False] * len(names)

    if allow_nones is not None: assert_list_types(allow_nones, (bool, int), "allow_nones")
    else: allow_nones = [None] * len(names)

    if eval_expr is not None: assert_list_types(names, str, "names")
    else: eval_expr = [None] * len(names)

    # Check lengths
    if (allow_nones is not None) and (must_be_present is not None):
        assert len(names) == len(types) == len(allow_nones) == len(must_be_present) == len(eval_expr), \
            f"Expected `{len(names)=}`, `{len(types)=}`, `{len(must_be_present)=}`, `{len(allow_nones)=}` and `{len(eval_expr)=}` to be of the same length"
    assert len(dict_to_check) <= len(names), f"Expected `len(dict_to_check) <= len(names)` but found `{len(dict_to_check)} <= {len(names)}`"

    # Check dict
    for name, type_, must_be, allow_none, eval_expr_ in zip(names, types, must_be_present, allow_nones, eval_expr):
        try:
            value = dict_to_check[name]
            assert_type(value, type_, f"{dict_name}[{name}]", bool(allow_none))
            if eval_expr_ is not None:
                assert_eval(value, eval_expr_, allow_none=True, name_of_check=name)
        except KeyError:
            assert must_be==False, f"Expected to find key:`{name}` in `{dict_name}={dict_to_check}`"


def assert_eval(to_check, eval_str: str, allow_none:bool=False, name_of_check:str=None):
    assert_types([eval_str, allow_none, name_of_check], [str, bool, str], ["eval_str", "allow_none", "name_of_check"], [0, 0, 1])
    if (to_check is None) and (allow_none is True):
        return

    assert "x" in eval_str, f"`{eval_str=}` must contain at least 1 'x' e.g. '0<=x<=1e6'"

    evaluation_exp = None
    try:
        evaluation_exp = eval_str.replace("x", str(to_check))
        success = eval(evaluation_exp)
    except Exception as e:
        raise RuntimeError(f"`Failed to execute the evaluation: {evaluation_exp=}` with the following error code:\n{e}")

    if not name_of_check:
        assert success, f"Evaluation failed: `{evaluation_exp}`"
    else:
        assert success, f"Evaluation of `{evaluation_exp}` with `{name_of_check}={to_check}` failed"


def assert_in(to_check, check_in, variable_name:str):
    """
    Check if the value `to_check` is present in `check_in`
    :param to_check: Value to be checked
    :param check_in: Values `to_check` is being checked against
    :param variable_name: Name which will be display with the assert fail
    """
    try:
        is_in = to_check in check_in
    except Exception:
        raise RuntimeError(f"Failed to execute `{to_check} in {check_in}`")

    if not is_in:
        raise ValueError(f"Expected `{variable_name}={to_check}` to be present in `{check_in}`")

def quick_assert_valid_image(image_source, advise:str= "\nTry using setting `parse_image=True` or parse it yourself with `parse_image_as_uint8_rgb_numpy_array(image_source)`"):
    if not isinstance(image_source, ndarray):
        raise ValueError("`image_source` is not a numpy array. Expected a uint8 numpy image of shape (height, width, 3), but received something else." + advise)
    if len(image_source.shape) != 3:
        raise ValueError("`image_source` has an invalid shape. Expected a uint8 numpy image of shape (height, width, 3), but received something else." + advise)
    h, w, c = image_source.shape
    image_seems_ok = all([isinstance(image_source, np.ndarray), (image_source.dtype == "uint8"), (c == 3), (h > 3), (w > 3)])
    if not image_seems_ok:
        raise ValueError("`image_source` is not valid. Expected a uint8 numpy image of shape (height, width, 3), but received something else." + advise)

def assert_color(color:tuple, color_type:str, variable_name:str) -> None:
    """
    >> assert_color((200, 20, 200, 1.0), "RGBA")
    """
    assert_types([color, color_type, variable_name], [tuple, str, str], ["color", "color_type", "variable_name"])
    if color_type == "RGB":
        wrong_number_of_channels = len(color) != 3
        if wrong_number_of_channels:
            raise ValueError(f"`{variable_name}={color}` must have exactly 3 channels (RGB)")
        not_valid_rgb = not all((isinstance(c, int) and 0 <= c <= 255) for c in color)
        if not_valid_rgb:
            raise ValueError(f"`{variable_name}={color}` is not a legal RGB color. Expect a tuple with exactly 3 number between 0<=channel<=255")
    elif color_type == "RGBA":
        wrong_number_of_channels = (len(color) != 4)
        if wrong_number_of_channels:
            raise ValueError(f"`{color=}` must have exactly 4 channels (RGBA)")
        not_valid_rgb = not all((isinstance(c, int) and 0 <= c <= 255) for c in color[:3])
        not_valid_alpha = not (0.0<=color[3]<=1.0)
        if not_valid_rgb or not_valid_alpha:
            raise ValueError(f"`{variable_name}={color}` is not a legal RGBA color. Expect a tuple with exactly 4 number.\n"
                             "The first 3 should be RGB values in 0<=channel<=255 while the 4'th should be an alpha value in 0.0<alpha<1.0")
    else:
        raise RuntimeError(f"`{variable_name}={color}` is either not implemented or unexpected")


def assert_valid_universal_inputs(
        title:Optional[str]=None,
        display_image: bool = True,
        return_image: bool = True,
        max_output_image_size_wh: tuple = None,
        save_image_path:Optional[str] = None,
        resize_factor: float = 1.0,
        BGR2RGB: bool = False,
        parse_image:bool=True
) -> None:
    # Checks - types
    assert_types(
        [title, display_image, return_image, max_output_image_size_wh, save_image_path, resize_factor, BGR2RGB, parse_image],
        [str,   bool,          bool,         tuple,                    str,             float,         bool,    bool],
        ["title", "display_image", "return_image", "max_output_image_size_wh", "save_image_path", "resize_factor", "BGR2RGB", "parse_image"],
        [1,     0,             0,            1,                        1,               0,             0,       0]
    )

    # `title`
    if isinstance(title, str) and (len(title) == 0):
        msg = f"You have passed an empty '{title=}'. This will probably cause some unwanted behavior"
        warnings.warn(msg)

    # `resize_factor`, `parse_image`
    assert_resize_factor_is_ok(resize_factor)
    if (resize_factor != 1.0) and (not parse_image):
        warning_msg = f"Found non default `{resize_factor=}` together with `{parse_image=}`. " \
                      f"Can only resize when `parse_image=True`. Will avoid parsing by setting `resize_factor=1.0` "
        warnings.warn(warning_msg)

    # `display_image`, `return_image`, `save_image_path`
    if not any([display_image, return_image, save_image_path]):
        raise ValueError("If you don't want to display, return or save the image. What's the point :)")
    if save_image_path is not None:
        formats = global_config.image_formats
        assert any(save_image_path.lower().endswith(f) for f in formats),\
            f"`{save_image_path=}` has an unknown image format. Please use one of these: `{formats}`"

    # `max_output_image_size_wh`
    if max_output_image_size_wh is not None:
        assert len(max_output_image_size_wh) == 2, f"Expected `{max_output_image_size_wh=}` to contain exactly 2 numbers (width and height). None is allowed"
        w, h = max_output_image_size_wh
        w_is_not_none, h_is_not_none = (w is not None), (h is not None)

        if (w_is_not_none and w < 10) or (h_is_not_none and h < 10):
            raise ValueError(f"Expected both `{max_output_image_size_wh=}` to contain two numbers larger than 10 pixels.")
        if (w_is_not_none and w < 100) or (h_is_not_none and h < 100):
            msg = f"`{max_output_image_size_wh=}` is suspiciously small. Are you sure it's correct?"
            warnings.warn(msg)
        elif (w_is_not_none and w > 10_000) or (h_is_not_none and h > 10_000):
            msg = f"`{max_output_image_size_wh=}` is suspiciously large. Are you sure it's correct?"
            warnings.warn(msg)

def assert_positive_float(to_check:float, zero_allowed:bool=False, max_value_allowed:float=None, variable_name:str=None) -> None:
    assert_types([to_check, zero_allowed, max_value_allowed, variable_name],
                 [float, bool, float, str],
                 ["to_check", "zero_allowed", "max_value_allowed", "variable_name"],
                 [0, 0, 1, 1])

    assert (max_value_allowed is None) or (max_value_allowed >= 0.0), f"Expected max_value >= 0, but found `{max_value_allowed=}`"
    if max_value_allowed is None:
        max_value_allowed = np.inf

    if (not zero_allowed) and ( (to_check <= 0.0) or (to_check > max_value_allowed) ):
        variable_name_string = f"{variable_name}=" if (variable_name is not None) else ""
        raise ValueError(f"`{variable_name_string}{to_check}` does not fulfill `0.0 < {to_check} <= {max_value_allowed}`")
    elif zero_allowed and ( (to_check < 0.0) or (to_check > max_value_allowed) ):
        variable_name_string = f"{variable_name}=" if (variable_name is not None) else ""
        raise ValueError(f"`{variable_name_string}{to_check}` does not fulfill `0.0 <= {to_check} <= {max_value_allowed}`")

def assert_positive_int(to_check:int, zero_allowed:bool=False, max_value_allowed:int=None, variable_name:str=None) -> None:
    assert_types([to_check, zero_allowed, max_value_allowed, variable_name],
                 [int, bool, int, str],
                 ["to_check", "zero_allowed", "max_value_allowed", "variable_name"],
                 [0, 0, 1, 1])

    assert (max_value_allowed is None) or (max_value_allowed >= 0), f"Expected max_value >= 0, but found `{max_value_allowed=}`"
    if max_value_allowed is None:
        max_value_allowed = np.inf

    if (not zero_allowed) and ( (to_check <= 0) or (to_check > max_value_allowed) ):
        variable_name_string = f"{variable_name}=" if (variable_name is not None) else ""
        raise ValueError(f"`{variable_name_string}{to_check}` does not fulfill `0 < {to_check} <= {max_value_allowed}`")
    elif zero_allowed and ( (to_check < 0) or (to_check > max_value_allowed) ):
        variable_name_string = f"{variable_name}=" if (variable_name is not None) else ""
        raise ValueError(f"`{variable_name_string}{to_check}` does not fulfill `0 <= {to_check} <= {max_value_allowed}`")

def assumed_torch_batch(image_source) -> bool:
    # TODO: This was written fast, make sure the check is sufficient
    if not TORCH_FOUND:
        return False
    if not isinstance(image_source, torch.Tensor):
        return False
    if not (len(image_source.shape) == 4):
        return False


    # In the following checks I expect either shape
    # (batch_size, channel, height, width) or
    # (batch_size, height, width, channel)
    if not (image_source.shape[3] == 3 or image_source.shape[1] == 3):
        return False
    if not ((image_source.shape[1] > global_config.image_min_height) or (image_source.shape[2] > global_config.image_min_height)):
        return False
    if not ((image_source.shape[2] > global_config.image_min_width) or (image_source.shape[3] > global_config.image_min_width)):
        return False
    return True

def assumed_numpy_batch(image_source:ndarray) -> bool:
    # TODO: This was written fast, make sure the check is sufficient
    if not isinstance(image_source, ndarray):
        return False
    if not (len(image_source.shape) == 4):
        return False
    if not (image_source.shape[3] == 3):
        return False
    if not (image_source.shape[1] > global_config.image_min_height):
        return False
    if not (image_source.shape[2] > global_config.image_min_width):
        return False
    return True

def assert_valid_video_path(video_path:str) -> None:
    assert_path(video_path, "video_path")
    image_formats  = global_config.image_formats
    video_formats_all = global_config.video_formats_all
    video_formats_tested = global_config.video_formats_tested

    if any(video_path.endswith("." + image_format) for image_format in image_formats):
        raise ValueError(f"You have provided a string with an image extension: `{video_path}`. Legal video formats are:\n`{video_formats_all}`")
    elif not any(video_path.lower().endswith("." + video_format) for video_format in video_formats_all):
        raise ValueError(f"Unrecognized video format in `{video_path}`. Legal video formats are:\n`{video_formats_all}`")
    elif not any(video_path.lower().endswith("." + video_format) for video_format in video_formats_tested):
        video_format = video_path.split(".")[-1]
        warning_msg = f"The file `{video_path}` has a valid video extension: `{video_format}`. " \
                      f"However, this format has not been thoroughly tested, and unexpected errors may occur. " \
                      f"The currently tested video formats are:\n{video_formats_tested}"
        warnings.warn(warning_msg)

def assert_resize_factor_is_ok(resize_factor):
    assert_type(resize_factor, float, "resize_factor")
    if resize_factor <= 0.0:
        raise ValueError(f"`{resize_factor=}` is not valid, it must be strictly greater than zero")
    if resize_factor > 100.0:
        warning_msg = f"Encountered a suspiciously large value for `{resize_factor=}`"
        warnings.warn(warning_msg)

__all__ = [
    "assert_path",
    "assert_image_type",
    "assert_valid_image",
    "is_valid_url",
    "assert_type",
    "assert_types",
    "assert_list_types",
    "assert_dict_types",
    "assert_eval",
    "assert_in",
    "quick_assert_valid_image",
    "assert_color",
    "assert_valid_universal_inputs",
    "assert_positive_float",
    "assert_positive_int",
    "assumed_torch_batch",
    "assumed_numpy_batch",
    "assert_valid_video_path",
    "assert_resize_factor_is_ok"
]


# TODO Remove this when debugging is done:
# TODO It's only here so i can import `justshowit` in jupyter notebook while also debugging in pycharm without having to change import statements to `from . _`
import sys as __sys
import os as __os
__sys.path.append(__os.path.abspath(__os.path.dirname(__file__)))

from .config import global_config
from .show import show
from .collage import show_collage
from .parsers import parse_torch_image_batch_as_uint8_rgb_numpy_array, parse_numpy_image_batch_as_uint8_rgb_numpy_array, parse_image_as_uint8_rgb_numpy_array, parse_arbitrary_image_source
from .grid import show_grid_configurable, show_grid
from .video import play_video, parse_video_to_images, parse_video_to_images_fixed_count, show_video
from .image_modifier import draw_text_cv2, draw_text_pillow, draw_image_title
from .utils import get_cv2_video_capture


###################################################################
# TODO - All around
###################################################################

# 20.) Add the possibility to save videos and individual frames with e.g. enter in `VideoPlayer`
# 21.) List torch as an optional package requirement
# 23.) `resize_universal_output_image` can be off by one pixel, fix it. Must be a simple rounded error somewhere
# 25.) Is it a good idea to have \n in errors? I'm thinking that it may fuck up the layout completely if used on a screen/terminal that is not standard size?
# 28.) Check all <NEW_VALUE> prints are correct had a very confusion problem with global_config.min_width and global_config.image_min_width
# 29.) Go through all the places i comment on comment names and check they are still valid
# 30.) Implement a random rotation option to show_collage
# 31.) Implement `play_audio`
# 32.) Implement `play_gif`
# 33.) add checks that the resized image in `resize_image_respect_aspect_ratio` is not below the minimum value defined in the global_config
# 34.) add resize to play_video perhaps even interactively
# 35.) the extra white space on the left handside of the text between these two: `row_text=["A"*100, "B", "C"], row_text=["A"*20, "B", "C"]` are not the same. They should be.
# 36.) Give a warning for `row_text=["A"*10000, "B", "C"]`
# 37.) setup(..., python_requires='>=3.??') what version should I set as a minimum for python, and should there be a minimum version for e.g. numpy as well?
# 38.) Change name on pip
# 39.) `max_output_image_size_wh` produce an image larger than the allowed in `show_collage`
# 40.) Should this be here, it gets triggered super often: "cv2's frame count is unreliable, this may be an indication that cv2 is struggling to read `video_path='/kaggle/input/video-file-for-lane-detection-project/test_video.mp4"
# 41.) Make a show_table function that can nicely display a pandas dataframe or a dict
# 42.) Make a no_parse option to play_video
# 43.) Make a get video info instead of having to use `return_video_details`
# 44.) This is not neccesary: return _parse_numpy(image) in `_parse_path` and it slows the program down by a lot
# 45.) Either cast a tuple to a list or throw and error when you recieve a tuple. Encountered some esoteric errors while testing tuples of image_sources
# 46.) I add to drop using the `from <module> import *` within the init function because the submodules wouldn't be visible to certain IDEs. It would be nice with a check that could ensure all functions, etc. is imported when e.g. something new is added
# 47.) Make a dedicated `show_torch_batch` function, which can reliably be used to plot batch_images

###################################################################
# #TODO Bugs
###################################################################

# Find out why this cannot be parsed correctly
# >> __image.just_show_image(np.ones((100, 200, 3), dtype="uint8") * (255, )*3)
# If the user provide a folder path that does not exists: throw an error e.g.
# >>> show(image, save_image_path="./asdjklæasdasdljks/IMAGE.png")

###################################################################
# TODO - Complete Rewrite
###################################################################
#
# 2.) checker.py                                          DONE
# 3.) collage.py                                          DONE
# 4.) parsers.py                                           DONE
# 5.) grid.py                                             DONE-ish (need tons more testing)
# 6.) video.py                                            DONE
# 7.) __image_modifier                                      DONE
# 8.) __show                                                DONE



from __future__ import annotations
from utils import get_cv2_video_capture
import cv2
import numpy as np
import time

class Slider:
    def __init__(self, image_width, image_height, max_value, padding=20, slider_height=20, add_frame_count:bool=True):
        max_amount_of_numbers = int(np.ceil(np.log10(max_value)))
        self.text_padding = max_amount_of_numbers * 10
        self.image_height = image_height
        self.image_width = image_width
        self.padding = padding
        self.slider_x = self.padding*2 + self.text_padding
        self.slider_width = image_width - self.slider_x - self.padding
        self.slider_height = slider_height
        self.slider_y = image_height - self.padding - self.slider_height
        self.indicator_radius = 10 if (image_width < 750) else 15
        self.indicator_radius_original = self.indicator_radius
        self.min_value = 0
        self.max_value = max_value
        self.slider_value = 0

        self.mouse_over_slider = False

        self.indicator_x = np.inf
        self.indicator_y = np.inf
        self.buffer = self.indicator_radius
        self.is_dragging = False
        self.max_indicator_multiplier = 1.5
        self.add_frame_count = add_frame_count

        self.color_slider_complete = tuple(reversed((230, 200, 150)))  # tuple(reversed( --> cv2 BGR shenanigans
        self.color_slider =    tuple(reversed((220, 220, 220)))
        self.color_indicator = tuple(reversed((254, 151, 0)))
        self.color_indicator_shadow = tuple(reversed((184, 111, 4)))
        self.color_text =      tuple(reversed((220, 220, 220)))
        self.color_text_box = tuple(reversed((40, 40, 40)))

    def is_mouse_over_slider(self, mouse_x, mouse_y):
        if not (mouse_x >= self.slider_x - self.buffer):
            return False
        elif not (mouse_x <= self.slider_x + self.slider_width + self.buffer):
            return False
        elif not (mouse_y >= self.slider_y - self.buffer):
            return False
        elif not (mouse_y <= self.slider_y + self.slider_height + self.buffer):
            return False
        return True

    def is_mouse_over_indicator(self, mouse_x, mouse_y):
        x, y, r = self.indicator_x, self.indicator_y, self.indicator_radius
        distance = ((mouse_x - x) ** 2 + (mouse_y - y) ** 2) ** 0.5
        r_buffed = (r + self.buffer)
        if distance <= r_buffed:
            denominator = max(1e-6, (distance/r_buffed))
            size_multiplier = max(min(self.max_indicator_multiplier, 1.0/denominator), 1.0)
            return True, size_multiplier
        return False, 1.0

    def update_slider_value(self, mouse_x):
        slider_value = round((mouse_x - self.slider_x) / self.slider_width * self.max_value)
        self.slider_value = max(0, min(self.max_value, slider_value))

    def handle_mouse_input(self, event, mouse_x, mouse_y):
        left_pressed_down = event == cv2.EVENT_LBUTTONDOWN
        left_released = event == cv2.EVENT_LBUTTONUP
        mouse_moved = event == cv2.EVENT_MOUSEMOVE
        self.mouse_over_slider = self.is_mouse_over_slider(mouse_x, mouse_y)

        if left_pressed_down and self.mouse_over_slider:
            if not self.is_dragging:
                self.update_slider_value(mouse_x)
            self.is_dragging = True
        elif left_released:
            self.is_dragging = False
        elif mouse_moved and self.is_dragging:
            self.update_slider_value(mouse_x)


        if self.is_dragging:
            self.indicator_radius = int(self.indicator_radius_original * self.max_indicator_multiplier)
        elif not self.is_dragging:
            mouse_over_indicator, size_multiplier = self.is_mouse_over_indicator(mouse_x, mouse_y)
            self.indicator_radius = int(self.indicator_radius_original * size_multiplier)
        else:
            self.indicator_radius = self.indicator_radius_original

    def draw_frame_count(self, image):
        # Settings
        text = str(self.slider_value+1)
        font = cv2.FONT_HERSHEY_DUPLEX
        scale = 1
        thickness = 1

        # Rectangle
        (text_width, text_height), _ = cv2.getTextSize(text, font, scale, thickness)
        text_x = 0
        text_y = text_height + 5
        cv2.rectangle(image, (text_x, 0), (text_x + text_width, text_y + 5), self.color_text_box, -1)

        # Draw the white text on top of the black box
        cv2.putText(image, text, (text_x, text_y), font, scale, self.color_text, thickness, cv2.LINE_AA)

    def draw_slider(self, image):
        # Draw frame count (upper left corner)
        if self.add_frame_count:
            self.draw_frame_count(image)

        # Draw frame count (next to play bar)
        text_p = (self.padding, self.slider_y + self.slider_height)
        cv2.putText(image, str(self.slider_value+1), text_p, 1, 1, self.color_text, 1, cv2.LINE_AA)

        # Draw slider bar (complete)
        rect_p1 = (self.slider_x, self.slider_y)
        rect_p2 = (self.slider_x + self.slider_width, self.slider_y + self.slider_height)
        cv2.rectangle(image, rect_p1, rect_p2, self.color_slider_complete, -1)

        # Draw slider bar
        self.indicator_x = int(self.slider_x + (self.slider_value / self.max_value) * self.slider_width)
        rect_p1 = (self.indicator_x, self.slider_y)
        rect_p2 = (self.slider_x + self.slider_width, self.slider_y + self.slider_height)
        cv2.rectangle(image, rect_p1, rect_p2, self.color_slider, -1)

        # Draw slider indicator
        self.indicator_y = self.slider_y + self.slider_height // 2
        cv2.circle(image, (self.indicator_x, self.indicator_y), self.indicator_radius+1, self.color_indicator_shadow, -1, lineType=cv2.LINE_AA)
        cv2.circle(image, (self.indicator_x, self.indicator_y), self.indicator_radius, self.color_indicator, -1, lineType=cv2.LINE_AA)


class VideoStream:
    def __init__(self, source:str | np.ndarray):
        self.is_cv2 = False
        self.video_path = None

        if isinstance(source, str):
            cap, info = get_cv2_video_capture(source)
            self.video_path = source
            self.is_cv2 = True
            self.source = cap
            self.info = info
            self.video_width =  info["video_width"]
            self.video_height = info["video_height"]
            self.total_frames = info["frame_count"]

        else:
            assert len(source.shape) == 4, f"Expected `source` to be of shape (frames, height, width, channels), but found {source.shape}"
            fn,h,w,c = source.shape
            assert c == 3, "Expected color images"
            self.source = source[:, :, :, ::-1]  # ::-1 is a fast RGB --> BGR
            self.video_width = w
            self.video_height = h
            self.total_frames = fn
        self.frame_index = 0

    def read(self):
        if self.is_cv2:
            successful, frame = self.source.read()
        else:
            frame = self.source[self.frame_index]
            successful = True
        self.frame_index += 1
        return successful, frame

    def release(self):
        if self.is_cv2:
            self.source.release()
        else:
            del self.source

    def set(self, flag, frame_index):
        if self.is_cv2:
            self.source.set(flag, frame_index)
            self.frame_index = self.source.get(cv2.CAP_PROP_POS_FRAMES)
        else:
            self.frame_index = frame_index


class VideoPlayer:
    def __init__(self, video_source:str | np.ndarray, add_frame_count:bool=True):
        # cv2
        self.cap = VideoStream(video_source)
        self.window_name = "Video Player"
        cv2.namedWindow(self.window_name)
        cv2.setMouseCallback(self.window_name, self.on_mouse, param=self)

        # video info
        self.video_path = self.cap.video_path
        self.video_width =  self.cap.video_width
        self.video_height = self.cap.video_height
        self.total_frames = self.cap.total_frames

        # Slider settings
        slider_height = 10
        padding = 20
        if (self.video_height > 800) or (self.video_width > 1000):
            padding = 30

        # General settings
        self.pad_size = (slider_height + padding * 2)
        self.color_pad_bottom = (100, 100, 100)
        self.min_video_width = 600
        self.side_pad = 0
        if self.video_width < self.min_video_width:
            self.side_pad = abs(self.video_width - self.min_video_width) // 2
        self.color_pad_sides = (20, 20, 20)

        # Mappings         escape       q                        a        left arrow    d         right arrow  1     2     3     4     5     6     7     8     9
        self.key_mapper = {113: "exit", 27: "exit", 32: "space", 97: "a", 2424832: "a", 100: "d", 2555904:"d", 49:1, 50:2, 51:3, 52:4, 53:5, 54:6, 55:7, 56:8, 57:9}
        self.fps_mapper =      {1: 1, 2: 3, 3: 5, 4: 15, 5: 30, 6: 60, 7: 120, 8:240, 9:999}
        self.fps_2_step_size = {v:k for k,v in self.fps_mapper.items()}

        # Adjustable variables
        self.current_frame = None
        self.current_frame_raw = None
        self.fps = self.fps_mapper[5]
        self.is_paused = False
        self.terminated = False

        # Adjust fps according to video length
        if self.total_frames <= 15:
            self.fps = self.fps_mapper[2]
        elif self.total_frames <= 30:
            self.fps = self.fps_mapper[3]
        elif self.total_frames <= 60:
            self.fps = self.fps_mapper[4]

        # Slider
        self.slider = Slider(
            self.video_width + self.side_pad * 2,
            self.video_height + self.pad_size,
            self.total_frames-1,
            padding,
            slider_height,
            add_frame_count
        )

        self.run()

    @staticmethod
    def on_mouse(event, x, y, _, params):
        self = params
        self.slider.handle_mouse_input(event, x, y)

    def pad_image(self, frame):
        if self.side_pad:
            frame = cv2.copyMakeBorder(frame, 0, 0, self.side_pad, self.side_pad, cv2.BORDER_CONSTANT, value=self.color_pad_sides)
        return cv2.copyMakeBorder(frame, 0, self.pad_size, 0, 0, cv2.BORDER_CONSTANT, value=self.color_pad_bottom)

    def next_frame(self, increment:bool):
        successful, frame = self.cap.read()
        assert successful, "This should always be successful, something unexpected has happend"
        self.current_frame_raw = frame.copy()
        self.current_frame = self.pad_image(frame)
        if increment and ((self.slider.slider_value-1) < self.total_frames):
            self.slider.slider_value += 1
        else:
            self.set_frame_index(self.slider.slider_value)

    def set_frame_index(self, frame_index:int):
        assert (frame_index >= 0) and (frame_index <= self.total_frames)
        self.cap.set(cv2.CAP_PROP_POS_FRAMES, frame_index)
        self.slider.slider_value = frame_index

    def clean_up(self):
        self.cap.release()
        cv2.destroyAllWindows()
        self.terminated = True

    def handle_key_input(self):
        # Extract and map keyboard inputs
        delay = int(1000 / self.fps) if not (self.is_paused or self.slider.is_dragging) else 5
        key_raw = cv2.waitKeyEx(delay)
        key = self.key_mapper.get(key_raw)
        close_button_pressed = cv2.getWindowProperty(self.window_name, cv2.WND_PROP_VISIBLE) < 1

        # Handle input
        current_frame_index = self.slider.slider_value
        frame_step_size = self.fps_2_step_size[self.fps]
        slowest_fps = list(self.fps_mapper.values())[0]
        frame_step_size_extra = 1 if (self.fps == slowest_fps and (not self.is_paused)) else 0

        if close_button_pressed or (key == "exit"):
            self.clean_up()
        elif key == "space":
            self.is_paused = not self.is_paused
        elif key == "a":
            frame_index = max(0, current_frame_index - (frame_step_size + frame_step_size_extra))
            self.set_frame_index(frame_index)
        elif key == "d":
            frame_index = min(self.total_frames-1, self.slider.slider_value + (frame_step_size - frame_step_size_extra))
            self.set_frame_index(frame_index)
        elif key in range(1, 10):
            self.fps = self.fps_mapper[key]

        if key in ["a", "d"]:
            cv2.waitKeyEx(1) # ensures that "a" and "d" gets shown as fast a possible
            return True, key
        elif key_raw != -1:
            time.sleep(1.0 / self.fps)
        return False, key

    def _run(self):
        while not self.terminated:
            manual_frame_skip_detected, key  = self.handle_key_input()
            if self.terminated:
                break

            # Restart video if end is reached while not paused or dragging
            if (self.slider.slider_value >= (self.total_frames-1)) and (not self.slider.is_dragging) and (not self.is_paused):
                self.set_frame_index(0)
                self.next_frame(increment=False)
            elif manual_frame_skip_detected:
                do_increment = not self.is_paused
                if (key == "a") and (self.slider.slider_value == 0):
                    do_increment = False
                self.next_frame(increment=do_increment) # TODO: rewrite `increment`, the reason it works is too convoluted
            elif self.is_paused and (not self.slider.is_dragging):
                self.current_frame = self.pad_image(self.current_frame_raw)
            elif self.slider.is_dragging:
                if self.slider.slider_value != self.cap.frame_index:
                    frame_index = max(0, min(self.total_frames-1, self.slider.slider_value))
                    self.set_frame_index(frame_index)
                self.next_frame(increment=False)
            elif (not self.is_paused) and (not self.slider.is_dragging):
                self.next_frame(increment=True)

            self.slider.draw_slider(self.current_frame)
            # TODO add call back for frame processing
            cv2.imshow(self.window_name, self.current_frame)

            if self.cap.frame_index != self.slider.slider_value: # TODO: rewrite, they should not have separate "shared" state
                self.cap.set(cv2.CAP_PROP_POS_FRAMES, self.slider.slider_value)

    def run(self):
        try:
            self._run()
        except Exception as e:
            self.clean_up()
            raise RuntimeError(f"Fail to show video with error:\n{e}")